@@ -7,25 +7,21 @@ echo "::group::Prepare repositories"
 
 os=$(cut -d: -f4 /etc/system-release-cpe)
 case ${os} in
-  amazon)
-    os=amazon-linux
-    major_version=$(cut -d: -f6 /etc/system-release-cpe)
-    DNF=yum
-    amazon-linux-extras install -y epel
-    ${DNF} install -y ca-certificates
-    ${DNF} install -y \
-           https://packages.groonga.org/${os}/${major_version}/groonga-release-latest.noarch.rpm
-    ;;
   almalinux|centos)
     major_version=$(cut -d: -f5 /etc/system-release-cpe | grep -o "^[0-9]")
     case ${major_version} in
       7)
         DNF=yum
         ;;
-      *)
+      8)
         DNF="dnf --enablerepo=powertools"
         ${DNF} module -y disable postgresql
         ;;
+      *)
+        DNF="dnf --enablerepo=crb"
+        ${DNF} install -y \
+               https://apache.jfrog.io/artifactory/arrow/${os}/${major_version}/apache-arrow-release-latest.rpm
+        ;;
     esac
 
     ${DNF} install -y \
@@ -45,14 +41,9 @@ echo "::group::Install built packages"
 
 packages_dir=/host/repositories/${os}/${major_version}/x86_64/Packages
 
-postgresql_version=$(basename $(ls ${packages_dir}/*-pgroonga-*.rpm | head -n1) | \
-                       grep -E -o '[0-9.]+' |
-                       head -n1)
-case ${os} in
-  amazon-linux)
-    amazon-linux-extras install -y postgresql${postgresql_version}
-    ;;
-esac
+pgroonga_package=$(basename $(ls ${packages_dir}/*-pgroonga-*.rpm | head -n1) | \
+                     sed -e 's/-pgroonga-.*$/-pgroonga/g')
+postgresql_version=$(echo ${pgroonga_package} | grep -E -o '[0-9.]+')
 
 ${DNF} install -y ${packages_dir}/*.rpm
 
@@ -62,14 +53,6 @@ echo "::endgroup::"
 echo "::group::Install packages for test"
 
 case ${os} in
-  amazon-linux)
-    if [ ${postgresql_version} -ne 13 ]; then
-      ${DNF} install -y libpq-devel
-    fi
-    ${DNF} install -y postgresql-server-devel
-    pg_config=pg_config
-    groonga_token_filter_stem_package_name=groonga-token-filter-stem
-    ;;
   almalinux|centos)
     case ${major_version} in
       7)
@@ -114,6 +97,8 @@ sudo -u postgres -H \
      --locale=C \
      --pgdata=${data_dir} \
      --username=root
+echo "max_prepared_transactions = 1" | \
+  sudo -u postgres -H tee --append ${data_dir}/postgresql.conf
 sudo -u postgres -H \
      $(${pg_config} --bindir)/pg_ctl start \
      --pgdata=${data_dir}
@@ -124,9 +109,11 @@ cp -a \
    /tmp/
 cd /tmp
 case "${os}" in
-  almalinux|amazon-linux|centos)
-    if [ ${postgresql_version} -lt 11 ]; then
-      rm sql/index-only-scan/include.sql
+  almalinux|centos)
+    if [ ${postgresql_version} -lt 12 ]; then
+      rm sql/function/highlight-html/declarative-partitioning.sql
+      rm sql/function/wal-set-applied-position/declarative-partitioning.sql
+      rm sql/vacuum/two-phase-commit.sql
     fi
     if [ ${postgresql_version} -lt 13 ]; then
       rm sql/full-text-search/text/single/declarative-partitioning.sql
@@ -170,6 +157,27 @@ if [ ${pg_regress_status} -ne 0 ]; then
   exit ${pg_regress_status}
 fi
 
+
+echo "::group::Upgrade"
+
+${DNF} remove -y ${pgroonga_package}
+
+# Disable upgrade test for first time packages.
+case ${postgresql_version} in
+  16) # TODO: Remove this after 3.1.4 release.
+    exit
+    ;;
+esac
+
+${DNF} install -y ${pgroonga_package}
+createdb upgrade
+psql upgrade -c 'CREATE EXTENSION pgroonga'
+${DNF} install -y ${packages_dir}/*.rpm
+psql upgrade -c 'ALTER EXTENSION pgroonga UPDATE'
+
+echo "::endgroup::"
+
+
 echo "::group::Postpare"
 
 sudo -u postgres -H \
