@@ -5,22 +5,70 @@ require "stringio"
 require "tempfile"
 
 module Helpers
+  class CommandRunError < StandardError
+    attr_reader :commane_line
+    attr_reader :output
+    attr_reader :error
+    def initialize(command_line, output, error)
+      @command_line = command_line
+      @output = output
+      @error = error
+      message = "failed to run: "
+      message << command_line.join(" ")
+      message << "\n"
+      message << "output:\n"
+      message << output
+      message << "error:\n"
+      message << error
+      super(message)
+    end
+  end
+
   module CommandRunnable
-    def spawn_process(*args)
+    def spawn_process(*command_line)
       env = {
         "LC_ALL" => "C",
         "PGCLIENTENCODING" => "UTF-8",
       }
-      output_read, output_write = IO.pipe
-      error_read, error_write = IO.pipe
-      options = {
-        :out => output_write,
-        :err => error_write,
-      }
-      pid = spawn(env, *args, options)
-      output_write.close
-      error_write.close
-      [pid, output_read, error_read]
+      IO.pipe do |input_read, input_write|
+        input_write.sync = true
+        IO.pipe do |output_read, output_write|
+          IO.pipe do |error_read, error_write|
+            options = {
+              in: input_read,
+              out: output_write,
+              err: error_write,
+            }
+            pid = spawn(env, *command_line, options)
+            begin
+              input_read.close
+              output_write.close
+              error_write.close
+              yield(pid, input_write, output_read, error_read)
+            ensure
+              begin
+                Process.waitpid(pid, Process::WNOHANG)
+              rescue SystemCallError
+                # Finished
+              else
+                Process.kill(:KILL, pid)
+                Process.waitpid(pid)
+              end
+            end
+          end
+        end
+      end
+    end
+
+    def read_command_output_all(input, initial_timeout: 1)
+      all_output = ""
+      timeout = initial_timeout
+      loop do
+        break unless IO.select([input], nil, nil, timeout)
+        all_output << read_command_output(input)
+        timeout = 0
+      end
+      all_output
     end
 
     def read_command_output(input)
@@ -34,46 +82,50 @@ module Helpers
       end
     end
 
-    def run_command(*args)
-      pid, output_read, error_read = spawn_process(*args)
-      output = ""
-      error = ""
-      status = nil
-      timeout = 1
-      loop do
-        readables, = IO.select([output_read, error_read], nil, nil, timeout)
-        if readables
-          timeout = 0
-          readables.each do |readable|
-            if readable == output_read
-              output << read_command_output(output_read)
-            else
-              error << read_command_output(error_read)
+    def run_command(*command_line)
+      spawn_process(*command_line) do |pid, input_write, output_read, error_read|
+        output = ""
+        error = ""
+        status = nil
+        timeout = 1
+        yield(input_write, output_read, output_read) if block_given?
+        loop do
+          readables, = IO.select([output_read, error_read], nil, nil, timeout)
+          if readables
+            timeout = 0
+            readables.each do |readable|
+              if readable == output_read
+                output << read_command_output(output_read)
+              else
+                error << read_command_output(error_read)
+              end
             end
+          else
+            timeout = 1
           end
-        else
-          timeout = 1
+          _, status = Process.waitpid2(pid, Process::WNOHANG)
+          break if status
         end
-        _, status = Process.waitpid2(pid, Process::WNOHANG)
-        break if status
+        output << read_command_output(output_read)
+        error << read_command_output(error_read)
+        unless status.success?
+          raise CommandRunError.new(command_line, output, error)
+        end
+        [output, error]
       end
-      output << read_command_output(output_read)
-      error << read_command_output(error_read)
-      unless status.success?
-        command_line = args.join(" ")
-        message = "failed to run: #{command_line}\n"
-        message << "output:\n"
-        message << output
-        message << "error:\n"
-        message << error
-        raise message
-      end
-      [output, error]
+    end
+  end
+
+  module PlatformDetectable
+    private
+    def windows?
+      /mingw|mswin|cygwin/.match?(RUBY_PLATFORM)
     end
   end
 
   class PostgreSQL
     include CommandRunnable
+    include PlatformDetectable
 
     attr_reader :dir
     attr_reader :host
@@ -81,6 +133,7 @@ module Helpers
     attr_reader :user
     attr_reader :replication_user
     attr_reader :replication_password
+    attr_reader :version
     def initialize(base_dir)
       @base_dir = base_dir
       @dir = nil
@@ -91,6 +144,8 @@ module Helpers
       @user = "pgroonga-test"
       @replication_user = nil
       @replication_password = nil
+      @version = nil
+      @pid = nil
       @running = false
     end
 
@@ -103,6 +158,7 @@ module Helpers
                port: 15432)
       @dir = File.join(@base_dir, db_path)
       @log_path = File.join(@dir, "log", @log_base_name)
+      @pgroonga_log_path = File.join(@dir, "pgroonga.log")
       socket_dir = File.join(@dir, "socket")
       @port = port
       @replication_user = "replicator"
@@ -132,6 +188,7 @@ module Helpers
       File.open(pg_hba_conf, "a") do |conf|
         conf.puts("host replication #{@replication_user} #{@host}/32 trust")
       end
+      @version = Integer(File.read(File.join(@dir, "PG_VERSION")).chomp, 10)
     end
 
     def create_replication_user
@@ -143,7 +200,7 @@ module Helpers
                   @replication_user)
     end
 
-    def init_replication(primary)
+    def init_replication(primary, shared_preload_libraries: [])
       @dir = File.join(@base_dir, "db-standby")
       @log_path = File.join(@dir, "log", @log_base_name)
       @port = primary.port + 1
@@ -162,7 +219,8 @@ module Helpers
       end
       conf = File.read(postgresql_conf)
       conf = conf.gsub(/^shared_preload_libraries = '(.*?)'/) do
-        "shared_preload_libraries = '#{$1},pgroonga_wal_applier'"
+        libraries = [$1, *shared_preload_libraries].join(",")
+        "shared_preload_libraries = '#{libraries}'"
       end
       File.write(postgresql_conf, conf)
     end
@@ -173,7 +231,7 @@ module Helpers
                     "-w",
                     "-D", @dir)
       rescue => error
-        error.message << "\nPostgreSQL log:\n#{File.read(@log_path)}"
+        error.message << "\nPostgreSQL log:\n#{read_log}"
         raise
       end
       loop do
@@ -187,25 +245,49 @@ module Helpers
         end
       end
       @running = true
+      pid_path = File.join(@dir, "postmaster.pid")
+      if File.exist?(pid_path)
+        first_line = File.readlines(pid_path, chomp: true)[0]
+        begin
+          @pid = Integer(first_line, 10)
+        rescue ArgumentError
+        end
+      end
     end
 
     def stop
       return unless running?
-      run_command("pg_ctl", "stop",
-                  "-D", @dir)
+      begin
+        run_command("pg_ctl", "stop",
+                    "-D", @dir)
+      rescue CommandRunError
+        if @pid
+          Process.kill(:KILL, @pid)
+          @pid = nil
+          @running = false
+        end
+        raise
+      else
+        @pid = nil
+        @running = false
+      end
     end
 
-    def psql(db, sql)
-      output, error = run_command("psql",
-                                  "--host", @host,
-                                  "--port", @port.to_s,
-                                  "--username", @user,
-                                  "--dbname", db,
-                                  "--echo-all",
-                                  "--no-psqlrc",
-                                  "--command", sql)
-      output = normalize_output(output)
-      [output, error]
+    def psql(db, *sqls, may_wait_crash_safer_preparing: false, &block)
+      if may_wait_crash_safer_preparing
+        begin
+          return psql_internal(db, *sqls, &block)
+        rescue Helpers::CommandRunError => error
+          case error.error.chomp
+          when "ERROR:  pgroonga: pgroonga_crash_safer is preparing"
+            # This may be happen on slow environment
+            sleep(3)
+          else
+            raise
+          end
+        end
+      end
+      psql_internal(db, *sqls, &block)
     end
 
     def groonga(*command_line)
@@ -216,9 +298,33 @@ module Helpers
       JSON.parse(output)
     end
 
+    def read_log
+      return "" unless File.exist?(@log_path)
+      File.read(@log_path)
+    end
+
+    def read_pgroonga_log
+      return "" unless File.exist?(@pgroonga_log_path)
+      File.read(@pgroonga_log_path)
+    end
+
     private
-    def windows?
-      /mingw|mswin|cygwin/.match?(RUBY_PLATFORM)
+    def psql_internal(db, *sqls, &block)
+      command_line = [
+        "psql",
+        "--host", @host,
+        "--port", @port.to_s,
+        "--username", @user,
+        "--dbname", db,
+        "--echo-all",
+        "--no-psqlrc",
+      ]
+      sqls.each do |sql|
+        command_line << "--command" << sql
+      end
+      output, error = run_command(*command_line, &block)
+      output = normalize_output(output)
+      [output, error]
     end
 
     def normalize_output(output)
@@ -236,6 +342,7 @@ module Helpers
 
   module Sandbox
     include CommandRunnable
+    include PlatformDetectable
 
     class << self
       def included(base)
@@ -255,20 +362,20 @@ module Helpers
       end
     end
 
-    def psql(db, sql)
-      @postgresql.psql(db, sql)
+    def psql(db, *sqls, **options, &block)
+      @postgresql.psql(db, *sqls, **options, &block)
     end
 
-    def run_sql(sql)
-      psql(@test_db_name, sql)
+    def run_sql(*sqls, **options, &block)
+      psql(@test_db_name, *sqls, **options, &block)
     end
 
-    def psql_standby(db, sql)
-      @postgresql_standby.psql(db, sql)
+    def psql_standby(db, *sqls, **options, &block)
+      @postgresql_standby.psql(db, *sqls, **options, &block)
     end
 
-    def run_sql_standby(sql)
-      psql_standby(@test_db_name, sql)
+    def run_sql_standby(*sqls, **options, &block)
+      psql_standby(@test_db_name, *sqls, **options, &block)
     end
 
     def groonga(*command_line)
@@ -317,9 +424,18 @@ module Helpers
     def teardown_db
     end
 
+    def shared_preload_libraries_standby
+      []
+    end
+
     def setup_standby_db
+      @postgresql_standby = nil
+      omit("TODO: Support Windows") if windows?
       @postgresql_standby = PostgreSQL.new(@tmp_dir)
-      @postgresql_standby.init_replication(@postgresql) do |conf|
+      options = {
+        shared_preload_libraries: shared_preload_libraries_standby,
+      }
+      @postgresql_standby.init_replication(@postgresql, **options) do |conf|
         conf.puts(additional_standby_configurations)
       end
       @postgresql_standby.start
@@ -333,6 +449,14 @@ module Helpers
       @postgresql_standby.stop if @postgresql_standby
     end
 
+    def start_postgres_standby
+      @postgresql_standby.start
+    end
+
+    def stop_postgres_standby
+      @postgresql_standby.stop
+    end
+
     def setup_reference_db
       @postgresql_reference = PostgreSQL.new(@tmp_dir)
       options = {
